@@ -1,49 +1,104 @@
-"""A sub-module to run a :class:`~thtools.core.ToeholdTest` across temperature ranges.`"""
+"""
+An implementation of running a :class:`~thtools.core.ToeholdTest` across temperature ranges.
+"""
 
-from typing import Sequence, Optional, Generator
+# This file is part of ToeholdTools (a library for the analysis of
+# toehold switch riboregulators created by the iGEM team City of
+# London UK 2021).
+# Copyright (c) 2021 Lucas Ng
+
+# ToeholdTools is free software: you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+
+# ToeholdTools is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+
+# You should have received a copy of the GNU General Public License
+# along with ToeholdTools.  If not, see <https://www.gnu.org/licenses/>.
+
+
+from typing import Sequence, Optional, Generator, List, Mapping, Union
 import datetime
+import re
+import logging
 
 import nupack
 import numpy as np
+import scipy.interpolate
 import prettytable
-import seaborn as sns
-import matplotlib
-import matplotlib.pyplot as plt
+import pandas as pd
+
+try:
+    import matplotlib
+    import matplotlib.pyplot as plt
+    import seaborn as sns
+except ImportError:
+    logging.warning(
+        "Seaborn not installed; please download it from https://pypi.org/project/seaborn/ if you wish to use graph plotting functionality."
+    )
 
 from . import CPU_COUNT
 from .core import ToeholdTest, ToeholdResult
 
 
+__all__ = ["CelsiusRangeTest", "CelsiusRangeResult"]
+
+
 class CelsiusRangeResult:
     """
     Data container for a :class:`CelsiusRangeTest`.
 
     Not intended to be constructed manually.
 
-    .. note:: All float values are given as decimals, not percentages.
+    .. note::
+        All activation and unbinding values are
+        given as decimals, not percentages.
 
     Attributes
     ----------
     results : Sequence[ToeholdResult]
-        The :class:`~thtools.core.ToeholdResult` objects from each temperature.
+        The :class:`~thtools.core.ToeholdResult` objects
+        from each temperature.
     celsius_range : Sequence[float]
+        The temperature range tested.
     targets : Sequence[str]
-        The highest activating item from the trigger_sets of each :class:`~thtools.core.ToeholdResult`.
+        The highest activating item from the trigger_sets
+        of each :class:`~thtools.core.ToeholdResult`.
     target_names : Sequence[str], optional
         The names of each item in :attr:`targets`.
+
+    activation : Sequence[float]
+        A list of the switch activation value when bound
+        to the target RNA as the temperature changes.
+    rbs_unbinding : Sequence[float]
+        A list of the RBS unbinding value when bound
+        to the target RNA as the temperature changes.
+    aug_unbinding : Sequence[float]
+        A list of the AUG unbinding value when bound
+        to the target RNA as the temperature changes.
+    post_aug_unbinding : Sequence[float]
+        A list of the post-AUG unbinding value when bound
+        to the target RNA as the temperature changes.
+    activation_se : Sequence[float]
+        A list of the switch activation's standard error
+        when bound to the target RNA as the temperature
+        changes.
+
+    specificity : Sequence[float]
+        A list of the switch specificity to the target
+        RNA as the temperature changes.
+    specificity_se : Sequence[float]
+        The standard error of the specificity.
+
     unix_created : datetime.datetime
+        The UNIX timestamp of the :class:`CelsiusRangeResult`.
     date : str
     meta : dict
     pretty_meta : str
-    activation : Sequence[float]
-    rbs_unbinding : Sequence[float]
-    aug_unbinding : Sequence[float]
-    post_aug_unbinding : Sequence[float]
-    specificity : Sequence[float]
-    activation_se : Sequence[float]
-        The standard error of the activation probability.
-    specificity_se : Sequence[float]
-        The standard error of the specificity.
 
     Examples
     --------
@@ -57,7 +112,7 @@ class CelsiusRangeResult:
         >>> rbs = "AGAGGAGA"
         >>> fasta = tt.FParser.fromspecies("Homo sapiens")[295:305]
         >>> triggers = fasta.seqs
-        >>> temperatures = range(10, 101, 10)
+        >>> temperatures = range(20, 101, 10)
         >>> my_test = tt.autoconfig(ths, rbs, triggers)
         >>> my_crt = tt.CelsiusRangeTest(my_test, temperatures)
         >>> my_result = my_crt.run(max_size=3, n_samples=200)
@@ -127,6 +182,21 @@ class CelsiusRangeResult:
         # assume the mode is the actual target
         self.inferred_target = max(set(self.targets), key=self.targets.count)
 
+    def __eq__(self, other):
+        return (
+            self.celsius_range == other.celsius_range
+            and self.results == other.results
+            and self._inferred_target == other._inferred_target
+        )
+
+    def __getitem__(self, key):
+        if isinstance(key, slice):
+            return self.__class__(self.results[key], self.celsius_range[key], self.meta)
+        elif isinstance(key, int):
+            return self.results[key]
+        else:
+            raise NotImplementedError("__getitem__ only supports slice and int")
+
     @property
     def inferred_target(self):
         """The target sequence inferred from the mode of the individual :class:`~thtools.core.ToeholdTest` highest activators."""
@@ -134,13 +204,14 @@ class CelsiusRangeResult:
 
     @inferred_target.setter
     def inferred_target(self, value):
-        self.meta["Target"] = self._inferred_target = value
         self.meta["Target name"] = self._inferred_target_name = self.target_names[
             self.targets.index(value)
         ]
+        self.meta["Target sequence"] = self._inferred_target = value
         self.pretty_meta = "\n".join(
-            [f"{key}: {value}" for key, value in self.meta.items()]
+            f"{key}: {value}" for key, value in self.meta.items()
         )
+
         self.specificity = []
         self.specificity_se = []
         self.activation = []
@@ -150,12 +221,10 @@ class CelsiusRangeResult:
         self.post_aug_unbinding = []
         for result in self.results:
             target_index = next(  # the first result from the below generator
-                (
-                    index
-                    for index, val in enumerate(result.trigger_sets)
-                    if val == self.inferred_target
-                    or (len(val) == 1 and val[0] == self.inferred_target)
-                )
+                index
+                for index, val in enumerate(result.trigger_sets)
+                if val == self.inferred_target
+                or (len(val) == 1 and val[0] == self.inferred_target)
             )
             self.activation.append(result.activation[target_index])
             self.activation_se.append(result.activation_se[target_index])
@@ -182,11 +251,15 @@ class CelsiusRangeResult:
     def __repr__(self):
         return f"<{self.__module__}.{type(self).__qualname__} of {len(self.results)} ToeholdTests at {hex(id(self))}>"
 
-    def plot(
-        self, y: str = "activation", filename: Optional[str] = None, swap: bool = False
-    ):
+    def copy(self):
+        """Copy an instance of CelsiusRangeResult."""
+        new = self.__class__(self.results, self.celsius_range, self.meta)
+        new.inferred_target = self.inferred_target
+        return new
+
+    def plot(self, y: str = "activation", z_score=1.96, swap: bool = False):
         """
-        Plot temperature against specificity, with color being the activation.
+        Plot temperature against activation, with color being the specificity.
 
         The following axes are plotted:
 
@@ -201,46 +274,86 @@ class CelsiusRangeResult:
               - temperature /°C
             * - y
               - toehold switch activation %
+                (or whatever :attr:`y` value you chose)
             * - color
               - specificity %
 
         Parameters
         ----------
-        filename : str, optional
-            The filename to save the figure to if desired.
+        y : str, default = "activation"
+            The attribute of the test to plot.
+            The options are:
+            - activation
+            - RBS unbinding
+            - AUG unbinding
+            - post-AUG unbinding
+        z_score : float, default = 1.96
+            The Z score to multiply SE with when drawing error bars.
+            The default represents a 95% confidence interval.
         swap : bool, optional
             Whether or not to swap the specificity and activation in the plot. Defaults to False.
         """
-        y = y.strip().lower().replace(" ", "_").replace("-", "_")
-        y_name = y.replace("_", " ").capitalize()
-        x_vals = self.celsius_range
+        y = re.sub(r"[\W ]+", "", y).strip().lower().replace(" ", "_")
+        y_name = y.replace("_", " ").capitalize() + " %"
+        x_vals = np.asarray(self.celsius_range)
         cmap = sns.color_palette("crest", as_cmap=True)
         fig, ax = plt.subplots()
         formatter = matplotlib.ticker.ScalarFormatter(useOffset=False)
         if not swap:
-            y_vals = np.array(getattr(self, y)) * 100
-            z_vals = np.array(self.specificity) * 100
-            scatter = ax.scatter(x_vals, y_vals, c=z_vals, cmap=cmap)
+            y_vals = np.asarray(getattr(self, y)) * 100
+            z_vals = np.asarray(self.specificity) * 100
+            y_err = np.asarray(self.activation_se) * z_score * 100
+            scatter = ax.scatter(x_vals, y_vals, c=z_vals, cmap=cmap, zorder=3)
             fig.colorbar(scatter, label="Specificity %", format=formatter)
             ax.set_xlabel("Temperature /°C")
             ax.set_ylabel(y_name)
         else:
-            y_vals = np.array(self.specificity) * 100
-            z_vals = np.array(getattr(self, y)) * 100
-            scatter = ax.scatter(x_vals, y_vals, c=z_vals, cmap=cmap)
+            y_vals = np.asarray(self.specificity) * 100
+            z_vals = np.asarray(getattr(self, y)) * 100
+            y_err = np.asarray(self.specificity_se) * z_score * 100
+            scatter = ax.scatter(x_vals, y_vals, c=z_vals, cmap=cmap, zorder=3)
             fig.colorbar(scatter, label=y_name, format=formatter)
             ax.set_xlabel("Temperature /°C")
             ax.set_ylabel("Specificity %")
+        ax.errorbar(
+            x_vals, y_vals, yerr=y_err, fmt="none", ecolor="black", capsize=3, zorder=1
+        )
+        self._lobf(ax, x_vals, y_vals, y_err)
         ax.set_axisbelow(True)
         ax.grid()
         plt.title(
-            f"CRT for toehold detecting {self.inferred_target_name if self.inferred_target_name is not None else self.inferred_target}"
+            f"CRT for toehold detecting {self.inferred_target_name or self.inferred_target}"
         )
-        if filename is not None:
-            fig.savefig(filename, dpi=1200)
-            return plt
-        else:
-            return plt
+
+        return plt
+
+    @staticmethod
+    def _lobf(ax, x_vals, y_vals, y_err: np.ndarray, k=3, zorder=2):
+        weights = np.divide(
+            1, y_err, out=np.full_like(y_err, 100), where=y_err != 0
+        )  # set w to 100 if ZeroDivision
+        # weights = y_err.max() - y_err
+        x_smooth = np.linspace(x_vals.min(), x_vals.max(), num=500)
+        y_smooth = scipy.interpolate.UnivariateSpline(x_vals, y_vals, w=weights, k=k)(
+            x_smooth
+        )
+        ax.plot(x_smooth, y_smooth, color="firebrick", zorder=zorder)
+
+    def savefig(
+        self,
+        fname,
+        y: str = "activation",
+        z_score=1.96,
+        swap: bool = False,
+        dpi=1200,
+        **kwargs,
+    ):
+        """
+        Convenience function saving the figure from :meth:`plot`.
+
+        See :meth:`plot`. All kwargs are passed to :func:`matplotlib.pyplot.savefig`.`
+        """
+        self.plot(y, z_score, swap).savefig(fname, dpi=dpi, **kwargs)
 
     def __str__(self):
         return self.prettify()
@@ -256,10 +369,13 @@ class CelsiusRangeResult:
         """
         table = prettytable.PrettyTable()
         table.add_column("Temperature /°C", self.celsius_range)
-        table.add_column("Target", self.targets)
-        if len(self.target_names) > 0 and not all(self.target_names):
-            names = ["+".join(name) for name in self.target_names]
-            table.add_column("Name", names)
+        if self.target_names and None not in self.target_names:
+            names = [
+                "+".join(name) if hasattr(name, "__iter__") else name
+                for name in self.target_names
+            ]
+            table.add_column("Target name", names)
+        table.add_column("Target sequence", self.targets)
         table.add_column("Activation %", [i * 100 for i in self.activation])
         table.add_column("Specificity %", [i * 100 for i in self.specificity])
         table.add_column("Activation SE", [i * 100 for i in self.activation_se])
@@ -324,6 +440,21 @@ class CelsiusRangeResult:
         """
         return self.tabulate(dp).get_json_string(**kwargs)
 
+    def to_df(self, dp: Optional[int] = None, **kwargs) -> pd.DataFrame:
+        """
+        Get the result as a pandas DataFrame.
+
+        Parameters
+        ----------
+        dp : int, optional
+            The decimal places to limit the display to.
+        **kwargs
+            Extra arguments to be passed to the
+            :class:`pandas.DataFrame` instance.
+        """
+        table = self.tabulate(dp)
+        return pd.DataFrame(table._rows, columns=table.field_names, **kwargs)
+
 
 ################################################################################
 
@@ -353,6 +484,8 @@ class CelsiusRangeTest:
     ----------
     thtest : ToeholdTest
     celsius_range : Sequence[float]
+    result : CelsiusRangeResult
+    meta : Mapping[str, str | int | float]
 
     Examples
     --------
@@ -374,6 +507,8 @@ class CelsiusRangeTest:
 
     thtest: ToeholdTest
     celsius_range: Sequence[float]
+    result: CelsiusRangeResult
+    meta: Mapping[str, Union[str, int, float]]
 
     def __init__(self, thtest: ToeholdTest, celsius_range: Sequence[float]):
         self.thtest = thtest
@@ -436,23 +571,25 @@ class CelsiusRangeTest:
         -----
         Unlike :meth:`generate`, this method has no significant performance penalty compared to :meth:`run`.
         """
-        chunks = []
+        chunks: List[ToeholdTest] = []
         for celsius in self.celsius_range:
             chunks.append(
                 self._adjust_temperature(celsius).run(max_size, n_samples, n_nodes)
             )
             yield chunks[-1]
-        self.result = CelsiusRangeResult(
-            chunks, self.celsius_range, meta={"THS": self.thtest.ths}
-        )
+        self.meta = chunks[-1].meta
+        del self.meta["Temperature /°C"]
+        del self.meta["Specificity %"]
+        del self.meta["Specificity SE"]
+        del self.meta["Runtime /s"]
+        self.result = CelsiusRangeResult(chunks, self.celsius_range, meta=self.meta)
 
     def _adjust_temperature(self, celsius: float) -> ToeholdTest:
         new_thtest = self.thtest.copy()
-        ensemble = self.thtest.model.ensemble
-        material = self.thtest.model.material
-        conditions = self.thtest.model.conditions.save()
-        del conditions["temperature"]
-        self.thtest.model = nupack.Model(
-            ensemble=ensemble, material=material, celsius=celsius, **conditions
+        model_params = new_thtest.model.to_json().to_object()
+        model_params["conditions"]["temperature"] = 273.15 + celsius
+        model_params["parameters"]["temperature"] = 273.15 + celsius
+        new_thtest.model = new_thtest.model.from_json(
+            nupack.JSON.from_object(model_params)
         )
         return new_thtest
